import unittest

from src.exceptionite import Handler, Tab
from src.exceptionite.exceptions import ConfigurationException


class OtherContextTab(Tab):
    id = "context"
    name = "An other context tab"


class TestHandler(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.handler = Handler()

    def test_handler_can_provide_basic_exception_data(self):
        try:
            raise ValueError("Custom message")
        except Exception as exception:
            self.handler.start(exception)

        assert self.handler.message() == "Custom message"
        assert self.handler.exception() == "ValueError"
        assert self.handler.namespace() == "builtins.ValueError"
        assert self.handler.count() > 0

        frame = self.handler.stacktrace()[0]
        assert frame.index == 0
        assert frame.relative_file == "tests/test_handler.py"
        assert not frame.is_vendor
        assert frame.lineno == 19
        assert frame.offending_line == 19
        assert frame.method == "test_handler_can_provide_basic_exception_data"

    def test_cannot_override_tab_context(self):
        with self.assertRaises(ConfigurationException):
            self.handler.renderer("web").add_tabs(OtherContextTab)



DEFAULT_OPTIONS = {
    "options": {
        "editor": "vscode",
        "search_url": "https://www.google.com/search?q=",
        "links": {
            "doc": "https://docs.masoniteproject.com",
            "repo": "https://github.com/MasoniteFramework/masonite",
        },
        "stack": {"offset": 8, "shorten": True},
        "hide_sensitive_data": True
    },
    "handlers": {
        "context": True,
        "solutions": {"stackoverflow": False, "possible_solutions": True},
        "recommendations": {"packages_updates": {"list": ["exceptionite"]}},
    },
}



import unittest
from dotty_dict import dotty

from src.exceptionite import Handler, Tab


class CustomTestTab(Tab):
    id = "test"
    name = "Test tab"

    def build(self):
        return {"key": "value"}


class TestWebRenderer(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.handler = Handler()
        try:
            raise ValueError("Custom message")
        except Exception as exception:
            self.handler.start(exception)

    def test_enabled_tabs(self):
        web = self.handler.renderer("web")
        assert len(web.enabled_tabs()) == 3

    def test_build_page_context(self):
        context = dotty(self.handler.renderer("web").build())
        assert context.get("exception.type") == "ValueError"
        assert context.get("exception.namespace") == "builtins.ValueError"
        assert context.get("exception.message") == "Custom message"
        assert context.get("config")
        assert context.get("actions") == []
        context_tab = context.get("tabs")[0]
        assert context_tab.get("id") == "context"
        assert context_tab.get("name") == "Context"
        assert context_tab.get("has_content")
        assert not context_tab.get("has_sections")
        blocks = context_tab.get("blocks")
        assert len(blocks) == 3
        env_block = blocks[0]
        assert env_block.get("id") == "environment"
        assert env_block.get("name") == "System Environment"
        assert env_block.get("has_content")
        assert env_block.get("icon") == "TerminalIcon"
        assert env_block.get("data").get("Arch") == "64bit"

    def test_add_tabs(self):
        web = self.handler.renderer("web")
        web.add_tabs(CustomTestTab)
        assert len(web.enabled_tabs()) == 4

    def test_add_actions(self):
        pass

    def test_can_display_error_page(self):
        exception = ValueError("Custom message")
        self.handler.start(exception)

        assert "Custom message" in self.handler.render("web")
        assert "ValueError" in self.handler.render("web")



import pkg_resources

from ..Block import Block


class Packages(Block):
    id = "packages"
    name = "Installed Packages"
    icon = "PuzzleIcon"
    disable_scrubbing = True

    def build(self):
        packages = {}
        for package in pkg_resources.working_set:
            packages.update({package.key: package.version})
        return packages

    def has_content(self):
        return True



import re

from ..Block import Block
from .. import solutions


class PossibleSolutions(Block):

    id = "possible_solutions"
    name = "Possible Solutions"
    component = "PossibleSolutionsBlock"
    advertise_content = True
    disable_scrubbing = True
    empty_msg = "No solution found for this error."
    icon = "LightBulbIcon"

    def __init__(self, tab, handler, options):
        super().__init__(tab, handler, options)
        self.registered_solutions = []
        self.register(
            solutions.DictionaryUpdateSequence(),
            solutions.ClassMethodExists(),
            solutions.ClassModelMethodExists(),
            solutions.QueryDefaultValue(),
            solutions.GetAttributeObject(),
            solutions.NoModuleNamed(),
            solutions.Syntax(),
            solutions.ImportIssue(),
            solutions.Undefined(),
            solutions.WrongParameterCount(),
            solutions.WrongConstructorParameterCount(),
            solutions.ObjectNotCallable(),
            solutions.SubscriptableIssue(),
        )

    def build(self):
        possible_solutions = []
        for solution in self.registered_solutions:
            r = re.compile(solution.regex())
            if r.match(self.handler.message()):
                description = solution.description()
                title = solution.title()
                matches = [m.groupdict() for m in r.finditer(self.handler.message())]
                for code, replacement in matches[0].items():
                    description = description.replace(":" + code, replacement)
                    title = title.replace(":" + code, replacement)

                possible_solutions.append({"title": title, "description": description})

        return {
            "first": possible_solutions[0] if len(possible_solutions) > 0 else None,
            "solutions": possible_solutions,
        }

    def register(self, *solutions):
        self.registered_solutions += solutions
        return self

    def has_content(self):
        return len(self.data.get("solutions")) > 0



from collections import OrderedDict
import os
from jinja2 import Environment, PackageLoader
from typing import Type, TYPE_CHECKING, List, Callable


if TYPE_CHECKING:
    from ..Tab import Tab
    from ..Handler import Handler
    from ..Action import Action

from ..tabs import ContextTab, SolutionsTab, RecommendationsTab
from ..exceptions import ConfigurationException, TabNotFound
from ..Block import Block


class WebRenderer:
    """Renderer used to render exception as a beautiful extendable HTML error page which ease
    debugging."""

    reserved_tabs = ["context"]

    def __init__(self, handler: "Handler") -> None:
        self.handler = handler
        # error page interface options
        self.tabs: dict = OrderedDict()
        self.actions: dict = {}
        self.context: dict = {}
        # setup base interface
        self.add_tabs(ContextTab, SolutionsTab, RecommendationsTab)

    def build(self) -> dict:
        """Build the handled exception context to inject into the error page."""
        from .. import __version__

        enabled_tabs = self.enabled_tabs()
        return {
            "config": {
                **self.handler.options.to_dict(),
                "absolute_path": os.getcwd(),
                "version": __version__,
            },
            "exception": {
                "type": self.handler.exception(),
                "message": self.handler.message(),
                "namespace": self.handler.namespace(),
                "stacktrace": self.handler.stacktrace().reverse().serialize(),
            },
            "tabs": [self.handler.scrub_data(tab.serialize()) for tab in enabled_tabs],
            "actions": [action.serialize() for action in self.actions.values()],
        }

    def render(self) -> str:
        """Render the HTML error page."""
        self.data = self.build()
        path = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), "templates", "exceptionite.js"
        )
        with open(path, "r") as f:
            script = f.read()

        env = Environment(loader=PackageLoader("exceptionite", "templates"))

        template = env.get_template("exception.html")
        return template.render({"data": self.data, "script": script})

    def add_tabs(self, *tab_classes: Type["Tab"]) -> "Handler":
        """Register a tab in the HTML error page."""
        for tab_class in tab_classes:
            tab = tab_class(self.handler)
            if tab.id in self.reserved_tabs and tab.id in self.tabs:
                raise ConfigurationException(
                    f"exceptionite: {tab.id} is a reserved name. This tab can't be overriden."
                )
            self.tabs.update({tab.id: tab})
        return self

    def add_actions(self, *action_classes: Type["Action"]) -> "Handler":
        """Register an action in the HTML error page."""
        for action_class in action_classes:
            action = action_class(self.handler)
            self.actions.update({action.id: action})
        return self

    def tab(self, id: str) -> "Tab":
        """Get registered tab with the given id."""
        try:
            return self.tabs[id]
        except KeyError:
            raise TabNotFound(f"Tab not found: {id}")

    def enabled_tabs(self) -> List["Tab"]:
        """Get enabled tabs from options"""
        return [
            tab
            for tab in self.tabs.values()
            if self.handler.options.get("handlers").get(tab.id, True)
        ]

    def run_action(self, action_id: str, options: dict = {}) -> dict:
        """Run the given action with options if any"""
        action = self.actions.get(action_id)
        return action.run(options)

    def add_context(self, name: str, data: "dict|Callable", icon: str = None) -> "WebRenderer":
        """Quick shortcut method to add a context block into 'context' tab."""

        custom_block = Block
        custom_block.id = f"context.id.{name.lower()}"
        custom_block.name = name
        custom_block.icon = icon
        if callable(data):
            custom_block.build = data
        else:
            custom_block.build = lambda self: data
        self.tab("context").add_blocks(custom_block)
        return self



class Block:
    id = ""
    name = "Block Name"
    icon = ""
    component = "BasicBlock"
    disable_scrubbing = False
    empty_msg = "No content."
    has_sections = False

    def __init__(self, tab, handler, options={}):
        # tab which holds the block
        self.tab = tab
        self.handler = handler
        self.options = options
        self.data = {}
        assert self.id, "Block should declare an 'id' attribute !"

    def serialize(self):
        raw_data = self.build()
        self.data = self.handler.scrub_data(raw_data, self.disable_scrubbing)
        return {
            "id": self.id,
            "name": self.name,
            "icon": self.icon,
            "component": self.component,
            "data": self.data,
            "empty_msg": self.empty_msg,
            "has_content": self.has_content(),
            "has_sections": self.has_sections,
        }

    def build(self):
        raise NotImplementedError("block should implement build()")

    def has_content(self):
        return True



import sys
import traceback
from dotty_dict import dotty
from typing import Type, TYPE_CHECKING
from typing_extensions import Protocol

if TYPE_CHECKING:

    class Renderer(Protocol):
        handler: "Handler"

        def __init__(self, handler: "Handler") -> None:
            ...

        def render(self) -> str:
            """Render exception with the given renderer"""
            ...


from .StackTrace import StackTrace
from .renderers import WebRenderer, TerminalRenderer
from .options import DEFAULT_OPTIONS as DefaultOptions


class Handler:
    """Exceptionite handler used to handle exceptions and render them using the given renderer."""

    scrub_keywords = [
        "password",
        "passwd",
        "pwd",
        "secret",
        "key",
        "api_key",
        "apikey",
        "access_token",
        "credentials",
        "token",
    ]

    def __init__(
        self,
    ):
        self.renderers: dict["Renderer"] = {}
        self.options: dict = dotty(DefaultOptions)
        self.context = {}
        self.add_renderer("web", WebRenderer)
        self.add_renderer("terminal", TerminalRenderer)

    def set_options(self, options: dict) -> "Handler":
        """Configure exceptionite handler with given options."""
        # ensure options is a dict here, might already be a dotty dict
        self.options = dotty(dict(options))
        return self

    def add_renderer(self, name: str, renderer_class: Type["Renderer"]) -> "Handler":
        """Register a renderer to handle exceptions."""
        self.renderers.update({name: renderer_class(self)})
        return self

    def renderer(self, name: str) -> "Renderer":
        """Get the renderer with the given name."""
        return self.renderers[name]

    def add_context(self, name: str, data: dict) -> "Handler":
        self.context.update({name: data})
        return self

    def start(self, exception: BaseException) -> "Handler":
        """Start handling the given exception."""
        self._exception = exception
        self._type, self._value, self._original_traceback = sys.exc_info()
        traceback_exc = traceback.TracebackException(
            self._type, self._value, self._original_traceback, capture_locals=True
        )
        self._stacktrace = StackTrace(
            traceback_exc,
            self._exception,
            offset=self.options.get("options.stack.offset"),
            shorten=self.options.get("options.stack.shorten"),
            scrubber=self.scrub_data,
        )
        self._stacktrace.generate().reverse()
        return self

    # helpers
    def exception(self) -> str:
        """Get the handled exception name."""
        return self._exception.__class__.__name__

    def namespace(self) -> str:
        """Get the handled exception full namespace."""
        return self._exception.__class__.__module__ + "." + self.exception()

    def message(self) -> str:
        """Get the handled exception message."""
        return str(self._exception)

    def stacktrace(self) -> "StackTrace":
        """Get the handled exception stack trace object."""
        return self._stacktrace

    def count(self):
        return len(self._stacktrace)

    def render(self, renderer: str) -> str:
        """Render the handled exception with the given renderer."""
        return self.renderer(renderer).render()

    def add_scrub_keywords(self, keywords: list) -> "Handler":
        """Add new scrub keywords used to hide sensitive data."""
        self.scrub_keywords.extend(keywords)
        # ensure keywords are not duplicated
        self.scrub_keywords = list(set(self.scrub_keywords))
        return self

    def set_scrub_keywords(self, keywords: list) -> "Handler":
        """Override scrub keywords used to hide sensitive data."""
        self.scrub_keywords = keywords
        return self

    def scrub_data(self, data: dict, disable: bool = False) -> dict:
        """Hide sensitive data of the given dictionary if enabled in the options with
        'hide_sensitive_data' parameter."""
        if not self.options.get("options.hide_sensitive_data") or disable:
            return data
        scrubbed_data = {}
        for key, val in data.items():
            if not val:
                scrubbed_data[key] = val
                continue
            if isinstance(val, dict):
                scrubbed_data[key] = self.scrub_data(val, disable)
            else:
                # scrub entire value if key matches
                should_scrub = False
                for token in self.scrub_keywords:
                    if token.lower() in key.lower():
                        should_scrub = True
                if should_scrub:
                    scrubbed_val = "*****"
                else:
                    scrubbed_val = val
                scrubbed_data[key] = scrubbed_val
        return scrubbed_data



class Action:
    name = "Action Name"
    icon = ""
    id = ""
    component = ""

    def __init__(self, handler) -> None:
        self.handler = handler

    def run(self, options={}):
        # do something
        return "Action executed !"

    def serialize(self):
        return {
            "id": self.id,
            "name": self.name,
            "icon": self.icon,
            "component": self.component,
        }



# flake8: noqa: E501


class DictionaryUpdateSequence:
    def title(self):
        return "Problem title"

    def description(self):
        return "Looks like you are trying to update a dictionary but are actually using a set or a list to update the dictionary instead. Double check what you are passing into the update method"

    def regex(self):
        return r"^dictionary update sequence element #"


class ClassMethodExists:
    def title(self):
        return "Class Method Exists"

    def description(self):
        return (
            "Check the :method attribute exists on the ':class' class. If this is a class you made then check the file its in and see if the method exists."
            "If this is a third party class then refer to the documentation."
        )

    def regex(self):
        return r"^class  \'(?P<class>([\w]*))\' has no attribute (?P<method>(\w+))"
class ClassModelMethodExists:
    def title(self):
        return "Model method does not exist"

    def description(self):
        return (
            "Could not find the ':method' method on the model class. Please check spelling. If this is a method you expect to be on the builder class then check the ORM documentation"
        )

    def regex(self):
        return r"^class model \'(?P<class>([\w]*))\' has no attribute (?P<method>(\w+))"

class QueryDefaultValue:
    def title(self):
        return "Missing default value for ':field'" 

    def description(self):
        return (
            "Default values are typically set on the database level. "
            "You can either add a default value on the :field table column in a migration or you should pass a value when creating this record"
        )

    def regex(self):
        return r"\(1364\, \"Field \'(?P<field>([\w]*))\' doesn't have a default value\"\)"


class GetAttributeObject:
    def title(self):
        return "Class Method Exists"

    def description(self):
        return """
            Double check the object you are using and make sure it has the :attribute attribute.

            If you are using a builtin python type then check Python documentation.
            If you are using your own class then check the available methods.
            """

    def regex(self):
        return r"^'(?P<object>(\w+))' object has no attribute '(?P<attribute>(\w+))'"


class NoModuleNamed:
    def title(self):
        return "Module Not Found Error"

    def description(self):
        return "This is an import error. Check the file where you imported the ': module' module. Make sure its spelled right and make sure you pip installed this module correctly if this is supposed to come from a PYPI package."

    def regex(self):
        return r"No module named '(?P<module>(\w+))'"


class Syntax:
    def title(self):
        return "Syntax Error"

    def description(self):
        return "Syntax errors are usually simple to fix. Just find the place that has invalid Python syntax and fix it. A good place to look is in your :class file on line :line"

    def regex(self):
        return r"^invalid syntax \((?P<class>(\w+\.py))+, line (?P<line>(\w+))"


class ImportIssue:
    def title(self):
        return "Import Issue"

    def description(self):
        return "This is an import error. Check the file where you imported the ':object' class and make sure it exists there."

    def regex(self):
        return r"^cannot import name '(?P<object>(\w+))'"


class Undefined:
    def title(self):
        return "Undefined Variable"

    def description(self):
        return "You are trying to use a variable that cannot be found. Check the ':variable' variable and see if it is declared, imported or in the correct scope depending on what the variable is."

    def regex(self):
        return r"name '(?P<variable>(\w+))' is not defined"


class WrongParameterCount:
    def title(self):
        return "Wrong Parameter Count"

    def description(self):
        return (
            "You have the wrong amount of parameters for the ':object' object. "
            "It requires :correct parameters but you gave :wrong parameters. If the parameters are stored in a variable try checking the variable to the left. "
            "If you are passing variables in normally then check the signature of the object"
        )

    def regex(self):
        return r"^(?P<object>(\w*))\(\) takes (?P<correct>(\d+)) positional (argument|arguments) but (?P<wrong>(\d+)) (were|was) given"


class WrongConstructorParameterCount:
    def title(self):
        return "Wrong Parameters to a Constructor"

    def description(self):
        return (
            "The ':object' object doesn't take parameters but you gave some anyway. "
            "Check the constructor of the ':object' object. It's likely it does not take any parameters. "
            "If its stored in a variable you can check the value to the left."
        )

    def regex(self):
        return r"^(?P<object>(\w*))\(\) takes no parameters "


class ObjectNotCallable:
    def title(self):
        return "Objects Cannot Be Called"

    def description(self):
        return (
            "You cannot call objects. The ':object' object has already been instiatiated. "
            "Once an object is instantiated it cannot be called directly anymore. "
            "Check if the ':object' is instantiated already."
        )

    def regex(self):
        return r"^'(?P<object>(\w*))' object is not callable"


class SubscriptableIssue:
    def title(self):
        return "Object Not Subscriptable"

    def description(self):
        return "Looks like you expected ':object' to be an iterable but it is not. You can only use subscrptions, like x[0], on iterable type objects (like lists, dicts, and strings) but not ':object' in this case."

    def regex(self):
        return r"^'(?P<object>(\w+))' object is not subscriptable"



from ..Tab import Tab
from ..blocks.Environment import Environment
from ..blocks.Packages import Packages
from ..blocks.Git import Git


class ContextTab(Tab):

    name = "Context"
    id = "context"
    icon = "ViewListIcon"

    def __init__(self, handler):
        super().__init__(handler)
        self.add_blocks(Environment, Packages, Git)



from ..Tab import Tab
from ..blocks.StackOverflow import StackOverflow
from ..blocks.PossibleSolutions import PossibleSolutions


class SolutionsTab(Tab):

    name = "Solutions"
    id = "solutions"
    icon = "LightBulbIcon"
    advertise_content = True

    def __init__(self, handler):
        super().__init__(handler)
        self.add_blocks(PossibleSolutions, StackOverflow)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..Handler import Handler


class JSONRenderer:
    """Renderer used to render exception as JSON payload."""

    def __init__(self, handler: "Handler") -> None:
        self.handler = handler
        self.data: dict = {}

    def build(self):
        return {
            "exception": {
                "type": self.handler.exception(),
                "namespace": self.handler.namespace(),
            },
            "message": self.handler.message(),
            "stacktrace": self.handler.stacktrace().reverse().serialize_light(),
        }

    def render(self) -> str:
        """Render the JSON payload."""
        self.data = self.build()
        return self.data



import shlex
import subprocess

from ..Block import Block


class Git(Block):
    id = "git"
    name = "Git"
    icon = "ShareIcon"
    disable_scrubbing = True

    def build(self):
        git_version = subprocess.check_output(shlex.split("git --version")).strip()
        try:
            commit = subprocess.check_output(
                shlex.split("git rev-parse HEAD"), stderr=subprocess.STDOUT
            ).strip()
        except subprocess.CalledProcessError:
            # not a git repository
            return {
                "commit": "",
                "branch": "",
                "git_version": git_version.decode("utf-8"),
                "remote": "",
            }
        branch = subprocess.check_output(
            shlex.split("git rev-parse --abbrev-ref HEAD"), stderr=subprocess.STDOUT
        ).strip()
        try:
            remote = subprocess.check_output(
                shlex.split("git config --get remote.origin.url"), stderr=subprocess.STDOUT
            ).strip()
        except subprocess.CalledProcessError:
            remote = b""

        return {
            "commit": commit.decode("utf-8"),
            "branch": branch.decode("utf-8"),
            "git_version": git_version.decode("utf-8"),
            "remote": remote.decode("utf-8"),
        }

    def has_content(self):
        return True



import requests
import urllib
from ..Block import Block


class StackOverflow(Block):
    id = "stackoverflow"
    name = "Stack Overflow Answers"
    component = "StackOverflowBlock"
    tags = ["python"]
    api_url = "https://api.stackexchange.com/2.2/"
    empty_msg = "No solution found on Stack Overflow."
    disable_scrubbing = True

    def get_tags(self):
        return ";".join(self.tags)

    def build(self):
        query = urllib.parse.urlencode(
            {
                "order": "asc",
                "sort": "relevance",
                "q": self.handler.message(),
                "body": self.handler.message(),
                "accepted": True,
                "tagged": self.get_tags(),
                "site": "stackoverflow",
            }
        )

        try:
            response = requests.get(f"{self.api_url}search/advanced?{query}").json()
        except requests.exceptions.ConnectionError:
            return {}

        accepted_answer_ids = []

        if not response.get("items"):
            query = urllib.parse.urlencode(
                {
                    "order": "desc",
                    "sort": "relevance",
                    "intitle": self.handler.exception(),
                    "site": "stackoverflow",
                    "filter": "!-*jbN-(0_ynL",
                    "tagged": self.get_tags(),
                    "key": "k7C3UwXDt3J0xOpri8RPgA((",
                }
            )
            response = requests.get(f"{self.api_url}search/advanced?{query}").json()

        for question in response.get("items", []):
            if "accepted_answer_id" in question:
                accepted_answer_ids.append(str(question["accepted_answer_id"]))

        query = urllib.parse.urlencode(
            {"order": "desc", "sort": "activity", "site": "stackoverflow"}
        )
        answers = requests.get(
            f"{self.api_url}answers/{';'.join(accepted_answer_ids)}?{query}"
        ).json()

        return {"questions": response["items"], "answers": answers}

    def has_content(self):
        return len(self.data.get("questions", [])) > 0



import pprint
import inspect
import os
from typing import List

MARKERS = ["site-packages/", "src/masonite/"]


class StackFrame:
    """Model a frame in the stack trace."""

    def __init__(self, index, frame_summary, variables={}, offset=5, shorten=False):
        self.index = index
        self.file = frame_summary[0]
        self.relative_file = None
        self.is_vendor = False
        if shorten:
            rel_path = self.file.replace(f"{os.getcwd()}/", "")
            # check if frame is a vendor frame (from an external python package or masonite package
            # in development)
            for marker in MARKERS:
                index = rel_path.find(marker)
                if index != -1:
                    break

            if index != -1:
                self.is_vendor = True
                end_index = len("src/") if marker == "src/masonite/" else len(marker)
                self.relative_file = "~/" + rel_path[index + end_index :]  # noqa: E203
            else:
                self.relative_file = rel_path

        self.language = self.get_language(self.file)

        self.offset = offset
        self.lineno = frame_summary[1]
        self.offending_line = self.lineno
        self.parent_statement = frame_summary[2]
        self.statement = frame_summary[3]
        self.start_line = self.lineno - offset if (self.lineno - offset > 0) else 0
        self.end_line = self.lineno + offset
        self.variables = variables
        self.method = frame_summary.name if frame_summary.name != "<module>" else "__main__"

        with open(self.file) as fp:
            printer = pprint.PrettyPrinter(indent=4)

            self.file_contents = printer.pformat(fp.read()).split("\\n")[
                self.start_line : self.end_line  # noqa: E203
            ]

        formatted_contents = {}
        read_line = self.start_line + 1

        for content in self.file_contents:
            if self.language == "python":
                formatted_line = (
                    content.replace("'\n '", "")
                    .replace("'\n \"", "")
                    .replace("\"\n '", "")
                    .replace('"\n "', "")
                    .replace("""(\'""", "")
                    .replace("')", "")
                )
            else:
                formatted_line = content.replace("'\n '", "").replace("'\n \"", "")

            if self.statement in formatted_line:
                self.offending_line = read_line

            formatted_contents.update({read_line: formatted_line})
            read_line += 1
        # remove common higher indentation
        min_start_spaces = 1000
        for line in formatted_contents.values():
            if line != "":
                min_start_spaces = min(min_start_spaces, len(line) - len(line.lstrip(" ")))

        self.file_contents = {
            number: text[min_start_spaces:] for number, text in formatted_contents.items()
        }

    def get_language(self, file: str) -> str:
        """Resolve language from the file path."""
        if file.endswith(".py"):
            return "python"
        elif file.endswith(".html"):
            return "html"

        return "python"


class StackTrace:
    """Model a stack trace."""

    trace: List["StackFrame"]

    def __init__(self, traceback, exception, offset=5, shorten=False, scrubber=None) -> None:
        self.traceback = traceback
        self.exception = exception
        self.trace = []
        self.loop_index = 0
        # options
        self.offset = offset
        self.shorten = shorten
        self.scrubber = scrubber

    def generate(self) -> List["StackFrame"]:
        """Generate all required data from a given traceback."""
        traceback = []
        for index, tb in enumerate(self.traceback.stack):
            traceback.append(
                StackFrame(
                    index,
                    tb,
                    variables=tb.locals,
                    offset=self.offset,
                    shorten=self.shorten,
                )
            )

        if hasattr(self.exception, "from_obj"):
            frame_summary = [
                inspect.getsourcefile(self.exception.from_obj),
                inspect.getsourcelines(self.exception.from_obj)[1],
                "",
                "null",
            ]
            traceback.append(
                StackFrame(
                    len(traceback),
                    frame_summary,
                    variables={},
                    offset=self.offset,
                    shorten=self.shorten,
                )
            )

        self.trace = traceback
        return self.trace

    def __iter__(self):
        self.loop_index = 0
        return self

    def __next__(self):
        if self.loop_index < len(self.trace):
            result = self.trace[self.loop_index]
            self.loop_index += 1
            return result
        else:
            raise StopIteration

    def __len__(self):
        return len(self.trace)

    def __getitem__(self, index: int) -> "StackFrame":
        return self.trace[index]

    def reverse(self) -> "StackTrace":
        """Set stack frame in reversed order. Exception is located at the beginning."""
        self.trace.sort(key=lambda frame: frame.index, reverse=True)
        return self

    def unreverse(self) -> "StackTrace":
        """Set stack frame in normal order. Exception is located at the end."""
        self.trace.sort(key=lambda frame: frame.index, reverse=False)
        return self

    def first(self) -> "StackFrame":
        """Get first frame of the stack trace."""
        return self.trace[0]

    def serialize(self) -> dict:
        """Serialize all data from the stack trace."""
        stack_data = []
        for frame in self.trace:
            frame_data = {
                "index": frame.index,
                "no": frame.lineno,
                "file": frame.file,
                "relative_file": frame.relative_file,
                "is_vendor": frame.is_vendor,
                "language": frame.language,
                "start": frame.start_line,
                "content": frame.file_contents,
                "variables": self.scrubber(frame.variables),
                "method": frame.method,
            }
            stack_data.append(frame_data)

        return stack_data

    def serialize_light(self) -> dict:
        """Serialize some data from the stack trace, to get a compact representation."""
        stack_data = []
        for frame in self.trace:
            frame_data = {
                "index": frame.index,
                "file": frame.file,
                "line": frame.lineno,
                "statement": frame.parent_statement,
            }
            stack_data.append(frame_data)
        return stack_data



import pkg_resources
import requests

from ..Block import Block


def get_latest_version(name):
    r = requests.get(f"https://pypi.org/pypi/{name}/json")
    if r.status_code == 200:
        version = r.json()["info"]["version"]
        return version
    return None


class PackagesUpdates(Block):
    id = "packages_updates"
    name = "Packages to update"
    icon = "ArrowCircleUpIcon"
    component = "PackagesUpdatesBlock"
    empty_msg = "Selected packages are up to date !"
    disable_scrubbing = True

    def build(self):
        installed_packages = {
            package.key: package.version for package in pkg_resources.working_set
        }

        packages_to_check = self.options.get("list", ["exceptionite"])
        packages = {}
        if packages_to_check:
            for package_name in packages_to_check:
                current_version = installed_packages.get(package_name)
                latest_version = get_latest_version(package_name)
                if current_version != latest_version:
                    packages.update(
                        {
                            package_name: {
                                "current": installed_packages.get(package_name),
                                "latest": latest_version,
                            }
                        }
                    )

        return packages

    def has_content(self):
        return len(self.data.keys()) > 0



from ..Tab import Tab
from ..blocks.PackagesUpdates import PackagesUpdates


class RecommendationsTab(Tab):

    name = "Recommendations"
    id = "recommendations"
    icon = "CheckCircleIcon"
    advertise_content = True

    def __init__(self, handler):
        super().__init__(handler)
        self.add_blocks(PackagesUpdates)



import os
from setuptools import setup

meta = {}
with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "src", "exceptionite", "version.py"),
    "r",
) as f:
    exec(f.read(), meta)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="exceptionite",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=meta["__version__"],
    description="Exception Handling Made Easy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    # The project's main homepage.
    url="https://github.com/masoniteframework/exceptionite",
    # Author details
    author="The Masonite Community",
    author_email="joe@masoniteproject.com",
    # Choose your license
    license="MIT license",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Environment :: Web Environment",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        # List package on masonite packages website
        "Framework :: Masonite",
    ],
    # What does your project relate to?
    keywords="Masonite, Python, Development, Framework",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    package_dir={"": "src"},
    packages=[
        "exceptionite",
        "exceptionite.blocks",
        "exceptionite.renderers",
        "exceptionite.tabs",
    ],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["jinja2", "requests", "colorama", "dotty-dict", "typing-extensions"],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[test]
    extras_require={
        "test": [
            "black",
            "flake8",
            "coverage",
            "pytest",
            "pytest-cov",
            "twine>=1.5.0",
            "wheel",
        ],
    },
)



from colorama import init as init_colorama, Fore, Style
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..Handler import Handler


COLORS_PREFIX = {
    "error": Fore.RED + Style.BRIGHT,
    "info": Fore.CYAN + Style.BRIGHT,
    "success": Fore.GREEN + Style.BRIGHT,
}


def color_output(color: str, output: str) -> str:
    return f"{COLORS_PREFIX.get(color)}{output}" + "\033[0m"


def get_frame_line_output(frame, index: int = None) -> str:
    frame_method = color_output("info", f"{frame.method}()")
    frame_line = f"{frame.relative_file or frame.file}: L{frame.lineno} in {frame_method}"
    if index:
        frame_index_colored = index if frame.is_vendor else color_output("success", index)
        return f"{frame_index_colored} {frame_line}"
    return frame_line


class TerminalRenderer:
    """Renderer used to print an exception nicely in the terminal. It will render stack
    trace too."""

    def __init__(self, handler: "Handler") -> None:
        self.handler = handler

    def render(self) -> str:
        """Print exception and stack trace nicely in terminal."""
        init_colorama()

        # start printing exception to terminal
        print("")
        print("")
        colored_exception_type = color_output("error", self.handler.exception())
        print(f"  {colored_exception_type}: {self.handler.message()}")
        print("")

        # show most recent frame of the stack
        stacktrace = self.handler.stacktrace().reverse()
        first_frame = stacktrace.first()
        first_frame_code = first_frame.file_contents
        print(f"   {get_frame_line_output(first_frame)}")
        print("")

        # get highest line number to align numbers with padding
        max_no_len = len(str(max(first_frame_code.keys())))
        for lineno, code_line in first_frame_code.items():
            lineno_str = str(lineno)
            if len(lineno_str) < max_no_len:
                for space in range(max_no_len - len(lineno_str)):
                    lineno_str += " "
            code_line = code_line.replace("&nbsp;", " ")
            # highlight line where exception raised
            if lineno == first_frame.offending_line:
                print(color_output("error", f" > {lineno_str} | {code_line}"))
            else:
                print(f"   {lineno_str} | {code_line}")
        print("")

        # show other frames
        title = color_output("info", f"Stack Trace ({len(stacktrace)}):")
        print(f"   {title}")
        print("")
        for index, frame in enumerate(stacktrace):
            frame_index = str(index + 1)
            print(f"   # {get_frame_line_output(frame, frame_index)}")
        return ""



# flake8: noqa: E501
from .Environment import Environment
from .Git import Git
from .Packages import Packages
from .PackagesUpdates import PackagesUpdates
from .PossibleSolutions import PossibleSolutions
from .StackOverflow import StackOverflow



from dotty_dict import dotty
from typing import TYPE_CHECKING, Type, List

if TYPE_CHECKING:
    from .Block import Block


class Tab:

    name = "Tab Name"
    icon = ""
    id = ""
    component = "BasicTab"
    advertise_content = False
    disable_scrubbing = False

    def __init__(self, handler) -> None:
        self.handler = handler
        self.blocks = {}
        tab_options = self.handler.options.get(f"handlers.{self.id}", {})
        self.options = dotty({} if isinstance(tab_options, bool) else tab_options)
        assert self.id, "Tab should declare an 'id' attribute !"

    def add_blocks(self, *block_classes: Type["Block"]):
        """Register the given block(s) in the tab."""
        for block_class in block_classes:
            block_options = self.options.get(block_class.id, {})
            block_options = dotty({} if isinstance(block_options, bool) else block_options)
            block = block_class(self, self.handler, block_options)
            self.blocks.update({block.id: block})
        return self

    def block(self, id: str) -> "Block":
        """Get tab block with the given name."""
        return self.blocks.get(id)

    def active_blocks(self) -> List["Block"]:
        """Get active blocks list enabled in options."""
        active_blocks = []
        display_tab = self.handler.options.get(f"handlers.{self.id}", True)
        if display_tab is False:
            return active_blocks
        elif display_tab is True:
            return self.blocks.values()

        for block in self.blocks.values():
            display_block = display_tab.get(block.id, True)
            if display_block is False:
                continue
            active_blocks.append(block)
        return active_blocks

    def serialize(self) -> dict:
        """Serialize data from the tab."""
        raw_data = self.build()
        self.data = self.handler.scrub_data(raw_data, self.disable_scrubbing)
        return {
            "id": self.id,
            "name": self.name,
            "icon": self.icon,
            "component": self.component,
            "blocks": list(map(lambda b: b.serialize(), self.active_blocks())),
            "data": self.data,
            "advertise_content": self.advertise_content,
            "has_content": self.has_content(),
        }

    def build(self):
        return {}

    def has_content(self) -> bool:
        """Compute if tab contains content. By default this will check if any of the block have
        content."""
        return any(
            list(map(lambda b: b.has_content(), self.active_blocks())),
        )
